/*
  maxflow - a program for sensitive sequence alignment

  $Id: maxflow.cpp,v 1.3 2004/02/24 09:50:11 aheger Exp $

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <vector>
#include <map>
#include <set>
#include <list>
#include <algorithm>

#include <iterator>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <string>

#include <cstdio>
#include <time.h>

#include <alignlib.h>

#include <alignlib/Alignator.h>
#include <alignlib/HelpersAlignator.h>
#include <alignlib/Alignandum.h>
#include <alignlib/HelpersSequence.h>
#include <alignlib/AlignataIterator.h>
#include <alignlib/Alignata.h>
#include <alignlib/HelpersAlignata.h>

#include "maxflow.h"
#include "prefix.h"

// #define DEBUG

#define SEPARATOR '\t'
#define SEPARATOR_PREFIX '-'

// how much to read in one chunk
#define MAX_LINE_SIZE 10000

using namespace std;
using namespace maxflow;
using namespace AlignLib;

typedef MaxFlow::Tree< Index, Weight > ClusterTree;

//-------------------------------------> parameter parsing <----------------------------------
/* Global parameters that are can be set by command line arguments */
unsigned int param_loglevel = 2;

static std::string param_filename_links("");
static std::string param_filename_prefixes("");

static bool param_pairwise = false;
static unsigned int param_alignment_mode = 0;
static std::vector<char *> param_prefixes;

static Weight param_min_dot_score = 1;

static CALCTYPE param_gop = 0;
static CALCTYPE param_gep = 0;

static bool param_output_dots = false;
static bool param_output_pairs = false;

static CALCTYPE param_iterative_alignment = 0;

static bool param_use_weights = false;

static unsigned int param_dump_data = 0;


#include <unistd.h>

const char * my_progname = "maxflow";

static void print_version() {
  std::cout << "# " << my_progname << " Version $Id: maxflow.cpp,v 1.3 2004/02/24 09:50:11 aheger Exp $" << endl;
}    

static void usage()
{
  print_version();

  std::cout << "Usage: " << my_progname << "[OPTIONS] [prefix1 prefix2 [prefix...]] [< graph input] " << std::endl << std::endl;
  std::cout << "-V   print version and exit." << std::endl;
  std::cout << "-h   prints out this message." << std::endl;
  
  std::cout << "-v   loglevel [" << param_loglevel << "]" << std::endl;

  std::cout << "-g   input file of graph  [" << param_filename_links << "]" << std::endl;  
  std::cout << "-p   input graph is in compressed format [" << param_pairwise << "]" << std::endl;
  std::cout << "-w   use alignment/pair weights [" << param_use_weights << "]" << std::endl;
  
  std::cout << "-m   minimum dot score [" << (int)param_min_dot_score << "]" << std::endl;
  std::cout << "-i   input file of prefixes to align [" << param_filename_prefixes << "]" << std::endl;

  std::cout << "-d   dump data [" << param_dump_data << "]" << std::endl;
  std::cout << "	first bit is set: dump weighted edges. " << std::endl;
  
  std::cout << "-o   gap opening penalty [" << param_gop << "]" << std::endl;
  std::cout << "-e   gap elongation penalty [" << param_gep << "]" << std::endl;  
  
  std::cout << "-n # specific alignment mode [" << param_alignment_mode << "]" << std::endl;
  std::cout << "	given a list of prefixes: p1 p2 p3 p4 p5 p6.... " << std::endl;
  std::cout << "	0: align pairs p1-p2 p3-p4 p5-p6 " << std::endl;
  std::cout << "	1: align p1-p2 p1-p3 p1-p4 p1-p5 p1-p6" << std::endl;
  std::cout << "	2: align p1-p2,p3,p4,p5,p6 p2-p1,p3,p4,p5,p6 p3-p1,p2,... " << std::endl;  
  std::cout << "	3: align p1-all p2-all p3-all p4-all p5-all p6-all" << std::endl;
  std::cout << "	4: align all-all" << std::endl;      
  
  std::cout << "options that specify the output" << std::endl;
  std::cout << "-I   iterative alignment [" << param_iterative_alignment << "]" << std::endl;
  std::cout << "-D   output as dots [" << param_output_dots << "]" << std::endl;
  std::cout << "-P   output as pairs in compressed format [" << param_output_pairs << "]" << std::endl;
}


//------------------------------------------------------------
void DumpArguments()
{
  print_version();

  std::cout << "####################################################" << std::endl;
  
  std::cout << "# loglevel			  : " << param_loglevel << std::endl;
  std::cout << "# dump				  : " << param_dump_data << std::endl;
  std::cout << "# input file graph		  : " << param_filename_links << std::endl;
  std::cout << "# graph is compressed		  : " << (param_alignment_mode ? "yes" : "no") << std::endl;
  std::cout << "# using weighted pairs/alignments : " << (param_use_weights ? "yes" : "no") << std::endl;  

  std::cout << "# minimum dot score		  : " << (int)param_min_dot_score << std::endl;

  std::cout << "# gap opening penalty		  : " << param_gop << std::endl;
  std::cout << "# gap elongation penalty	  : " << param_gep << std::endl;  
  std::cout << "# iterative alignment		  : " << param_iterative_alignment << std::endl;

  std::cout << "# pairwise alignment mode	  : " << (param_pairwise ? "yes" : "no") << std::endl;
  std::cout << "# output dots			  : " << (param_output_dots ? "yes" : "no") << std::endl;
  std::cout << "# output compressed alignment	  : " << (param_output_pairs ? "yes" : "no") << std::endl;    

  std::cout << "# alignment mode		  : " << param_alignment_mode << std::endl;
  std::cout << "# prefixes to align		  : " << param_prefixes.size() << std::endl;
  
  if (param_loglevel >= 3)
    {
      std::cout << "#";
      for (unsigned int i = 0; i < param_prefixes.size(); ++i)
	std::cout << " " << param_prefixes[i];
      std::cout << std::endl;
    }

  std::cout << "####################################################" << std::endl;
  
}

void ParseArguments (int argc, char *argv[]) {

  int c;  

  extern char * optarg;

  while ((c=getopt(argc, argv, "V?hv:d:g:m:n:pi:DPI:e:o:w")) != EOF) {
    switch(c) {
    case 'v':
      param_loglevel = atoi(optarg); break;
    case 'd':
      param_dump_data = atoi(optarg); break;
    case 'g':
      param_filename_links = optarg; break;
    case 'i':
      param_filename_prefixes = optarg; break;
    case 'p':
      param_pairwise = true; break;
    case 'w':
      param_use_weights = true; break;
    case 'n':
      param_alignment_mode = atoi(optarg); break;
    case 'm':
      param_min_dot_score = (Weight)(atof(optarg)); break;
    case 'o':
      param_gop = atof(optarg); break;
    case 'e':
      param_gep = atof(optarg); break;
    case 'I':
      param_iterative_alignment = atof(optarg); break;
    case 'D':
      param_output_dots = true; break;
    case 'P':
      param_output_pairs = true; break;
    case 'V':
      print_version(); exit(EXIT_SUCCESS);
    case '?':
    case 'h':
      usage(); exit(EXIT_SUCCESS);
    }
  }

  // set pointers to end of options
  (argc)-=optind;
  (argv)+=optind;

  if (!(param_output_dots || param_output_pairs))
    {
      usage();
      std::cout << "please specify either output as dots (-D) or as pairs (-P)." << std::endl;
      exit(EXIT_FAILURE);
    }

  // read prefixes to align and put them into the param_prefixes array
  if (param_filename_prefixes.length() > 0)
    {
      FILE * file = fopen(param_filename_prefixes.c_str(),"r");

      if (file == NULL)
	{
	  std::cout << "# could not open input file " << param_filename_prefixes << std::endl;
	  exit(EXIT_FAILURE);
	}

      while (!feof( file ))
	{
	  Token token = new char[MAX_TOKEN_SIZE];
      
	  if (fscanf(file, "%s", token) == 1)
	    param_prefixes.push_back( token );
	  else 
	    delete [] token;
	}
      fclose( file );
    }
      
  while (argc > 0)
    {
      Token token = new char[MAX_TOKEN_SIZE];
      strncpy( token, argv[0], MAX_TOKEN_SIZE);
      param_prefixes.push_back( token );
      argv++;
      --argc;
    }

  if (param_alignment_mode == 0 && ((param_prefixes.size() % 2) != 0) && param_dump_data == 0)
    {
      std::cerr << "with alignment mode 0 an even number of prefixes has to be specified." << std::endl;
      exit(EXIT_FAILURE);
    }
  
  if (param_loglevel >= 2) DumpArguments();
  
}

//------------------------------------------------------------
char * GetPrefix( const char * in) {
  const char * pos = index(in, '-');
  if (pos == NULL) {
    std::cerr << "could not parse string " << in << endl;
    exit(1);
  }	 
  int l = pos - in;
  char * result = new char[ l + 1];
  strncpy( result, in, l);
  result[l] = '\0';
  return result;
}


// sort edges in descending order
struct NcmpLt : public binary_function<EdgeIndex, EdgeIndex, bool>
{
  NcmpLt( const Edges & edges ) : mEdges(edges) {};
  
  bool operator()(const EdgeIndex & e1, const EdgeIndex & e2) const
  {
    return mEdges[e1] > mEdges[e2];
  }

  const Edges & mEdges;
};

/** Sorts neighbours for each vertex by index.

    A defined sort order is necessary for the set
    comparison methods to work when weighting the edges.
 */
void SortNeighbours( Edges & edges, Neighbours & neighbours ) 
{
  typedef Neighbours::iterator NeighboursIterator;
  NcmpLt cmp( edges );

  // sort each neighbour list and remove duplicates
  {
    NeighboursIterator it(neighbours.begin()), end(neighbours.end());
    
    for (Vertex i = 0; it!= end; ++it, ++i) {
      EdgeVector tmp;
      
      //// add match to oneself
      // it->push_front( Neighbour( i, i) );
      
      std::copy( it->begin(), it->end(), back_inserter< EdgeVector >( tmp ));
      // note: sorting has to be done in reverse order, because elements
      // are added to slist from the front
      std::sort( tmp.begin(), tmp.end(), cmp);
      it->clear();
      std::unique_copy( tmp.begin(), tmp.end(), front_inserter< NeighbourList>(*it));
    }
  }
}

//------------------------------------------------------------
void BuildNeighbourListsFromDots( Edges & edges,
				  Neighbours & neighbours, 
				  MapToken2Vertex & map_token2vertex, 
				  MapVertex2Token & map_vertex2token,
				  FILE * file)
{

  typedef MapToken2Vertex::const_iterator MapIterator;
  typedef Neighbours::iterator NeighboursIterator;
  
  int iteration = 0;

  EdgeIndex edge_index = 0;

  char buffer[MAX_LINE_LENGTH];
  
  while(!feof(file))
    {

      iteration ++;
      if (iteration % 100000 == 0) 
	std::cout << "# iteration=" << iteration << std::endl;

      Token token1 = new char[MAX_TOKEN_SIZE];
      Token token2 = new char[MAX_TOKEN_SIZE];
      int x;
      
      Weight weight = 1;

      fgets(buffer, MAX_LINE_LENGTH, file);
      if (buffer[0] == '#') continue;
      if (sscanf(buffer, "%s\t%s\t%d", token1, token2, &x) != 3) break;
      
      if (param_use_weights)
	weight = x;

      // do not allow self-loops
      if (strcmp(token1,token2) == 0) continue;
      
      Index index1 = UpdateMapsDelete( map_token2vertex, map_vertex2token, token1);
      Index index2 = UpdateMapsDelete( map_token2vertex, map_vertex2token, token2);
      
      // std::cout << token1 << "=" << index1 <<" " << token2 << "=" << index2 << " " << weight << std::endl;
      
      if (index1 >= (Index)neighbours.size())
	neighbours.resize(index1+1);
      
      if (index2 >= (Index)neighbours.size())
	neighbours.resize(index2+1);

      Index min_index = index1, max_index = index2;

      if (index1 > index2)
	{
	  min_index = index2;
	  max_index = index1;
	}
      
      // check if neighbour has already been touched
      // by looking it up in the exisiting neighbour lists
      // is this faster than using a hash?
      // at least it conserved memory.
      // Note: this works, because mSource in edges is always
      // smaller than mTarget.
      {
	NeighbourList::iterator it(neighbours[min_index].begin()), end(neighbours[min_index].end());
	bool found = false;
	for (;it!=end;++it)
	  if (edges[*it].mTarget == max_index)
	    {
	      found = true;
	      break;
	    }
	if (found) continue;
      }
      
      edges.push_back( Edge(min_index, max_index, (Weight)weight) );
      
      neighbours[min_index].push_front(edge_index);
      neighbours[max_index].push_front(edge_index);

      ++edge_index;
    }

  SortNeighbours( edges, neighbours );
  
}


//-----------------------------------------------------------
  void BuildNeighbourListsFromAlignments( Edges & edges,
					  Neighbours & neighbours, 
					  MapToken2Vertex & map_token2vertex, 
					  MapVertex2Token & map_vertex2token, 
					  FILE * file )
  {

  typedef MapToken2Vertex::const_iterator MapIterator;
  typedef Neighbours::iterator NeighboursIterator;
  
  char query_ali[MAX_LINE_LENGTH];
  char sbjct_ali[MAX_LINE_LENGTH];

  char prefix1[MAX_TOKEN_SIZE];
  char prefix2[MAX_TOKEN_SIZE];

  Weight weight = 1;
  int query_from, query_to, sbjct_from, sbjct_to;

  AlignLib::Alignata * map_query2row = AlignLib::makeAlignataVector();

  if (param_loglevel >= 2)
      std::cout << "# reading graph (.=1000): ";

  EdgeIndex edge_index = 0;

  int iteration = 0;

  char buffer[MAX_LINE_LENGTH];
  
  while(!feof(file))
    {

      iteration ++;
      if (param_loglevel >= 2 && iteration % 1000 == 0)
	if (iteration % 10000 == 0)
	  std::cout << ":" << std::flush;	  
	else
	  std::cout << "." << std::flush;

      float x;

      fgets(buffer, MAX_LINE_LENGTH, file);
      if (buffer[0] == '#') continue;
      
      if (sscanf(buffer, "%s\t%s\t%f\t%i\t%i%s\t%i\t%i\t%s", 
		 prefix1, prefix2, &x, 
		 &query_from, &query_to, query_ali,
		 &sbjct_from, &sbjct_to, sbjct_ali
		 ) != 9) break;

      if (param_use_weights)
	weight = (int)x;
      else
	weight = 1;
	  
      // do not allow self-loops
      if (strcmp(prefix1,prefix2) == 0) continue;
      
    
      AlignLib::fillAlignataCompressed( map_query2row, 
					query_from, std::string(query_ali), 
					sbjct_from, std::string(sbjct_ali));
      
      AlignLib::AlignataIterator it(map_query2row->begin());
      AlignLib::AlignataIterator it_end(map_query2row->end());
      
      char * p1 = new char[MAX_TOKEN_SIZE];
      strcpy( p1, prefix1 );
      int l1 = strlen(prefix1);
      p1[l1++] = '-';
      
      char * p2 = new char[MAX_TOKEN_SIZE];
      strcpy( p2, prefix2 );
      int l2 = strlen(prefix2);
      p2[l2++] = '-';

      for (; it != it_end; ++it)
	{

	  Token token1 = new char[MAX_TOKEN_SIZE];
	  strncpy( token1, p1, l1);
	  sprintf( &token1[l1], "%i", it->mRow);
	  
	  Token token2 = new char[MAX_TOKEN_SIZE];
	  strncpy( token2, p2, l2);
	  sprintf( &token2[l2], "%i", it->mCol);
	  
	  Index index1 = UpdateMapsDelete( map_token2vertex, map_vertex2token, token1);
	  Index index2 = UpdateMapsDelete( map_token2vertex, map_vertex2token, token2);
	  
	  if (index1 >= (Index)neighbours.size())
	    neighbours.resize(index1+1);

	  if (index2 >= (Index)neighbours.size())
	    neighbours.resize(index2+1);

	  Index min_index = index1, max_index = index2;

	  if (index1 > index2)
	    {
	      min_index = index2;
	      max_index = index1;
	    }
	  
	  // check if neighbour has already been touched
	  // by looking it up in the exisiting neighbour lists
	  // is this faster than using a hash?
	  // at least it conserved memory.
	  // Note: this works, because mSource in edges is always
	  // smaller than mTarget.
	  {
	    NeighbourList::iterator it(neighbours[min_index].begin()), end(neighbours[min_index].end());
	    bool found = false;
	    for (;it!=end;++it)
	      if (edges[*it].mTarget == max_index)
		{
		  found = true;
		  break;
		}
	    if (found) continue;
	  }
	  
	  edges.push_back( Edge(min_index,max_index,weight) );
	  
	  neighbours[min_index].push_front(edge_index);
	  neighbours[max_index].push_front(edge_index);

	  ++edge_index;
	  
	}
      
      // skip to end of line
      fgets(query_ali, MAX_LINE_LENGTH, file);

      map_query2row->Clear();
      
    }

  if (param_loglevel == 2) std::cout << std::endl;

  delete map_query2row;

  SortNeighbours( edges, neighbours );  
}

//----------------------------------------------------------------------------
// calculate new edge weights
// edge-weight = number of common neighbours/number of total neighbours per node
// calculate clustering coefficients, too
void CalculateEdgeWeights( Edges & edges,
			   Neighbours & neighbours,
			   const MapVertex2Token & map_vertex2token )
  
{

  long num_vertices = map_vertex2token.size();
  std::vector< long > edges_in_neighbourhood( num_vertices, 0);

  {
    // iterate over all edges
    Edges::iterator it(edges.begin()), end(edges.end());

    if (param_loglevel == 2)
      std::cout << "# weighting graph (.=1000):";
    
    for (EdgeIndex edge_index = 0; it != end; ++it, ++edge_index) {      
      
      if (param_loglevel == 2 && edge_index % 1000 == 0)
	if (edge_index % 10000 == 0)
	  std::cout << ":" << std::flush;	  
	else 
	  std::cout << "." << std::flush;

      Vertex source = it->mSource;
      Vertex target = it->mTarget;

      unsigned int c = 0;  // size of common set
      unsigned int u = 0;  // size of unique set 
      unsigned int t = 0;  // size of total set
	
      {
	NeighbourList::const_iterator it(neighbours[source].begin()), iend( neighbours[source].end());
	NeighbourList::const_iterator jt(neighbours[target].begin()), jend( neighbours[target].end());	  

	while ( it != iend && jt != jend )
	  {
	    Weight wi = edges[*it].mPriority;
	    Weight wj = edges[*jt].mPriority;
	    
	    if ( edges[*it].getNeighbour(source) < edges[*jt].getNeighbour(target) )
	      {
		t += wi;
		u += wi;
		++it;
	      }
	    else if ( edges[*it].getNeighbour(source) == edges[*jt].getNeighbour(target) )
	      {
		// do not consider the edge linking source and target itself
		if (*it != *jt) {
		  t += (wi + wj) / 2;
		  c += (wi + wj) / 2;
		}
		++it;
		++jt;
	      }
	    else
	      {
		t += wj;
		u += wj;
		++jt;
	      }
	  }
	while (it != iend)
	  {
	    Weight wi = edges[*it].mPriority;	    
	    u+=wi;
	    t+=wi;
	    ++it;
	  }
	
	while (jt != jend)
	  {
	    Weight wj = edges[*jt].mPriority;
	    u+=wj;
	    t+=wj;
	    ++jt;
	  }
      }
      
      // add two to the common set for source and target
      // remove two from the uniqe set for source and target
      c += 2;
      u -= 2;

      if (param_loglevel >= 3)
	{
	  std::cout << "# "
		    << map_vertex2token[source] << '\t'
		    << map_vertex2token[target] << '\t'
		    << c << '\t'
		    << u << '\t'
		    << t - u -c << '\t'
		    << t << '\t'
		    << c * 100 / t 
		    << std::endl;
	}
      
      Weight weight = Weight(c * 100 / t);
      
      it->mWeight = weight;
      
    }
    
    if (param_loglevel == 2) std::cout << std::endl;
    
  }
  
}

//------------------------------------------------------------
void PrintNeighbours( const Edges & edges,
		      const Neighbours & neighbours,
		      const MapVertex2Token & map_vertex2token )
{
#ifdef DEBUG
  if (param_loglevel >= 4) {
    Neighbours::const_iterator it(neighbours.begin()), end(neighbours.end());
    for (Index i = 0; it != end; ++it, ++i) {
      if ((*it).size() == 1) continue;
      std::cout << "# " << i << "=" << map_vertex2token[i] << ":";
      NeighbourList::const_iterator iit( it->begin()), iend( it->end());
      for (;iit!=iend;++iit)
	std::cout << edges[*iit].getNeighbour(i) << "(" << (int)edges[*iit].mPriority << ")" << "-";
      std::cout << std::endl;
    }
  }
#endif
}

//------------------------------------------------------------
/* build clusters

   Method 1:
	sort all edges by weight and then use old algorithms:
	mst by calculating connected components

	Advantages:
		a. code is there
		b. tree can be build at the same time as the
			clustering.
		c. joint/disjoint test can be performed
			easily.

	Disadvantages:
		a. speed: sorting all edges will take time
		b. memory: an array of all edges is needed

   Method 2:
	use Kruskal (O(E log E)) or priority search (O(E*V log V))

	Advantages:
		a. not much extra memory

	Disadvantage:
		a. complicated, because of the prefixes, it is
		not a single tree, but many trees.
		b. theoretically correct? Taking always the closest
		neighbour is not correct, as the closest neighbour
		might be incompatible.

   ergo: I choose method 1
   
 */

// sort edges in descending order
struct EcmpGt : public binary_function<EdgeVector::const_iterator, EdgeVector::const_iterator, bool>
{
  EcmpGt( const Edges & edges ): mEdges(edges) {};
  
  bool operator()(const EdgeIndex s1, const EdgeIndex & s2)
  {
    return mEdges[s1].mWeight > mEdges[s2].mWeight;
  }

  const Edges & mEdges;
};

void CalculateClusters( const Edges & edges,
			const Neighbours & neighbours,
			const MapVertex2Index & map_vertex2prefix_index,
			MapVertex2Index & map_vertex2cluster,
			Clusters & clusters,
			ClusterTree & tree){

  int num_vertices = map_vertex2prefix_index.size();
  
  // create sorted list of all edges
  EdgeVector edge_indices( edges.size() );

  for (EdgeIndex e = 0; e < (EdgeIndex)edge_indices.size(); ++e) edge_indices[e] = e; 
  
  if (param_loglevel >= 2)
    std::cout << "# sorting edges" << std::endl;
  
  {
    EcmpGt cmp(edges);
    
    std::sort( edge_indices.begin(),
	       edge_indices.end(),
	       cmp );
  }
  
  // iterate through all edges and calculate minimum spanning tree + clusters

  EdgeVector::const_iterator it(edge_indices.begin()), end(edge_indices.end());
  
  // note: 0 is needed as dummy element, therefore arrays are larger than one.
  // the array is shrunk at the end.
  std::vector< PrefixSet > set_prefixes(num_vertices+1);
  map_vertex2cluster.clear();
  map_vertex2cluster.resize(num_vertices + 1, 0);

  unsigned long index = 0;

  if (param_loglevel == 2)
    std::cout << "# calculating mst (.=1000): ";
  
  for ( ; it != end; ++it)
    {
      if (param_loglevel == 2 && ++index % 1000 == 0)
	if (index % 10000 == 0)
	  std::cout << ":" << std::flush;	  
	else 
	  std::cout << "." << std::flush;

      Index index1 = edges[*it].mSource;
      Index index2 = edges[*it].mTarget;
      Weight weight = edges[*it].mWeight;
      
      Index prefix_index1 = map_vertex2prefix_index[index1];
      Index prefix_index2 = map_vertex2prefix_index[index2];      

      /* Sedgewick: Algo in C, p 507 */
      {
	int x,y,i,j,t;
	x = i = index1 + 1;
	y = j = index2 + 1;

	while (map_vertex2cluster[i] > 0) i = map_vertex2cluster[i];
	while (map_vertex2cluster[j] > 0) j = map_vertex2cluster[j];


	while (map_vertex2cluster[x] > 0) 
	  {
	    t = x; x = map_vertex2cluster[x]; map_vertex2cluster[t] = i; 
	  }

	while (map_vertex2cluster[y] > 0) 
	  {
	    t = y; y = map_vertex2cluster[y]; map_vertex2cluster[t] = j; 
	  }

	set_prefixes[i].insert( prefix_index1 );
	set_prefixes[j].insert( prefix_index2 );

	// if nodes belong to two different components join them,
	// but only if they contain disjoint sets of proteins.
	if (i != j)
	  {
	
	    // calculate set intersections of prefixes in components.
	    std::vector<Index>intersection;
	    std::set_intersection( set_prefixes[i].begin(), set_prefixes[i].end(),
				   set_prefixes[j].begin(), set_prefixes[j].end(),
				   std::back_insert_iterator< std::vector<Index> >(intersection));
	

	    
	    if (intersection.size() == 0) {

	      // add to tree
	      {
		Index x = tree.getLastParent( index1 );
		Index y = tree.getLastParent( index2 );
		tree.joinNodes( x, y, weight );
	      }
	      
	      // std::cout << map_vertex2token[index1] << SEPARATOR << map_vertex2token[index2] << SEPARATOR << weight << std::endl;
	  
	      if (map_vertex2cluster[j] < map_vertex2cluster[i])
		{
		  map_vertex2cluster[j] += map_vertex2cluster[i] -1; map_vertex2cluster[i] = j; 
		  std::copy(set_prefixes[i].begin(),
			    set_prefixes[i].end(),
			    std::inserter< PrefixSet >(set_prefixes[j], set_prefixes[j].begin()));
		}
	      else
		{
		  map_vertex2cluster[i] += map_vertex2cluster[j] -1; map_vertex2cluster[j] = i; 
		  std::copy(set_prefixes[j].begin(),
			    set_prefixes[j].end(),
			    std::inserter< PrefixSet >(set_prefixes[i], set_prefixes[i].begin()));
		}
	    }
	  }
      }
    }

  if (param_loglevel == 2) std::cout << std::endl;          
  
  {
    // rename clusters starting from 0
    
    clusters.clear();
    clusters.resize( num_vertices );
    
    std::vector< Index > map_old2new(num_vertices + 1, -1);
    int current_cluster = 0;
    
    for (Vertex v = 1; v <= num_vertices; ++v) {
      
      Index old_cluster = map_vertex2cluster[v];
      
      if (old_cluster <= 0) old_cluster = v;
      
      Index new_cluster = map_old2new[old_cluster];

      if (new_cluster < 0)
	{
	  new_cluster = current_cluster++;
	  map_old2new[old_cluster] = new_cluster;
	}
      
      map_vertex2cluster[v] = new_cluster;
      clusters[new_cluster].push_back( v - 1 );
      
    }
    
    // delete the first element from map_vertex2cluster
    map_vertex2cluster.erase(map_vertex2cluster.begin());

    // delete unnecessary elements from clusters
    clusters.resize( current_cluster );
  }
  
}

//------------------------------------------------------------
/* iterate through all vertices and build lists of prefixes
 */
void BuildPrefixMaps( Residues & residues,
		      MapPrefix2PrefixIndex & map_prefix2prefix_index,
		      MapPrefixIndex2Prefix & map_prefix_index2prefix,
		      MapVertex2Index & map_vertex2prefix_index,
		      const MapVertex2Token & map_vertex2token )
{
  
  for (unsigned int vertex = 0; vertex < map_vertex2token.size(); ++vertex)
    {
      char * prefix = GetPrefix(map_vertex2token[vertex]);
      Index prefix_index = UpdateMapsDelete( map_prefix2prefix_index, map_prefix_index2prefix, prefix);
      map_vertex2prefix_index.push_back( prefix_index );

      // map of prefix_index to vertex and back
      if (prefix_index >= (Index)residues.size())
	residues.resize( prefix_index+1 );
      
      residues[prefix_index].push_back(vertex);
    }
}


//----------------------------------------------------------------------------------
/* calculate score in a cluster
 */
inline Weight GetScoreCluster( const ClusterTree & tree, Index & path_start, Index & path_end)
{

  Index a,b;

  a = tree.getParent(path_start);
  b = tree.getParent(path_end);

  // find common node:
  while (!tree.isRoot(a) && !tree.isRoot(b)) {

    if (a == b)
      return tree.getWeight(a);
    
    if (a < b )
      a = tree.getParent(a);
    else
	b = tree.getParent(b);
    
  }

  if (a < b)
    return tree.getWeight(b);
  else
    return tree.getWeight(a);
  
}

//------------------------------------------------------------
inline Residue GetResidue( const char * in) {
  const char * pos = index(in, SEPARATOR_PREFIX);
  if (pos == NULL) {
    return atoi(in);
  } else {
    return atoi(pos+1);
  }
}

//------------------------------------------------------------
/* calculate dots
 */
void CalculateDots( Alignata * dots,
		    const Edges & edges,
		    const Neighbours & neighbours,
		    const char * prefix_query,
		    const char * prefix_sbjct,
		    const Residues & residues,
		    const ClusterTree & tree,
		    const Clusters & clusters,
		    const MapVertex2Index & map_vertex2cluster,
		    const MapVertex2Token & map_vertex2token,
		    const MapPrefix2PrefixIndex & map_prefix2prefix_index,
		    const MapVertex2Index & map_vertex2prefix_index,
		    const MapPrefixIndex2Prefix & map_prefix_index2prefix) 
{
  Index query, sbjct;
      
  { 
    MapPrefix2PrefixIndex::const_iterator it;
    
    if ((it = map_prefix2prefix_index.find( (Token)prefix_query )) != map_prefix2prefix_index.end())
      query = (*it).second;
    else {
      if (param_loglevel >= 1)
	std::cout << "# ERROR: prefix " << prefix_query << " not found." << std::endl;
      return;
    }
    
    if ( (it = map_prefix2prefix_index.find( (Token)prefix_sbjct) ) != map_prefix2prefix_index.end())
      sbjct = (*it).second;
    else {
      if (param_loglevel >= 1)
	std::cout << "# ERROR: prefix " << prefix_sbjct << " not found." << std::endl;
      return;
      
    }
  }
    
  int nquery = residues[query].size();
  int nsbjct = residues[sbjct].size();
  
  if (param_loglevel >= 2)
    std::cout << "# nquery(" << query << ":" << prefix_query << ")=" << nquery 
	      << "  nsbjct(" << sbjct << ":" << prefix_sbjct << ")=" << nsbjct 
	      << std::endl;

  //----------------------------------------------------------------------------  
  // build map of residues in sbjct to matrix column
  std::vector< Index > map_index2weights(*(std::max_element( residues[sbjct].begin(), residues[sbjct].end())) + 1, 0 );  
  std::vector< Index > map_weights2vertex( nsbjct );
  
  {
    ResidueList::const_iterator it(residues[sbjct].begin()), end(residues[sbjct].end());
    for (unsigned int i = 0; it!= end; ++it, ++i)
      {
	map_index2weights[*it] = i;
	map_weights2vertex[i] = *it;
      }
  }
    
#ifdef DEBUG
  if (param_loglevel >= 4) 
    PrintIndex( map_index2weights );
#endif  

  //----------------------------------------------------------------------------  
  // calculate dots = main alignment loop
  {
    
    unsigned int row = 0;
    
    ResidueList::const_iterator q_it( residues[query].begin()), q_end( residues[query].end());
    
    for (; q_it != q_end; ++q_it) {

      std::vector< Weight > weights(nsbjct,0);
      
      Index q = *q_it;
      
      if (param_loglevel >= 3) 
	std::cout << "# processing " << row << "/" << nquery << ": residue " << q << "=" << map_vertex2token[q] << std::endl;
      
      {
	Index qc = map_vertex2cluster[q];
	//----------------------------------------------------------
	// 1. 1st direction: q -> cluster -> graph -> t
	// iterate over all members of cluster that q is in
	Cluster::const_iterator i_it( clusters[qc].begin()), i_end( clusters[qc].end());
	for (; i_it != i_end; ++i_it) {
	  
	  Index i = *i_it;

	  Weight w1;

	  if (q != i) 
	    w1 = GetScoreCluster( tree, q, i);
	  else
	    w1 = MAX_WEIGHT;
	  
	  if (param_loglevel >= 4) 
	    std::cout << "# cluster: " << qc << " residue=" << map_vertex2token[i] << " weight=" << (int)w1 << std::endl;
	  
	  //----------------------------------------------------------
	  // i is already sbjct sequence
	  // matrix = max(matrix, w1 = h(s,i))
	  // and go to beginning as aggregate score can only be less
	  if (map_vertex2prefix_index[i] == sbjct) 
	    if (weights[map_index2weights[i]] < w1) {
	      weights[map_index2weights[i]] = w1;
	      continue;
	    }
	  
	  //----------------------------------------------------------	
	  // iterate over all edges from i in graph
	  // matrix = max(matrix, min( w1, w2 = g(i,t))
	  NeighbourList::const_iterator ei(neighbours[i].begin()), ei_end(neighbours[i].end());
	  
	  for(; ei != ei_end; ++ei)
	    {

	      Index t = edges[*ei].getNeighbour( i );	    
	      Weight w2 = edges[*ei].mWeight;
	    
	      if (map_vertex2prefix_index[t] == sbjct)
		{
	      
		  Weight w = std::min( w1, w2 );
	      
		  if (weights[map_index2weights[t]] < w)
		    weights[map_index2weights[t]] = w;

		  if (param_loglevel >= 4) 
		    std::cout << "#   graph: residue=" << map_vertex2token[t] 
			      << " weight=" << (int)w2 
			      << " sequence=" << map_prefix_index2prefix[map_vertex2prefix_index[t]]
			      << " total=" << (int)w		  
			      << std::endl;
		}
	    }
	}
      }

      {
	//----------------------------------------------------------	
	// iterate over all edges from q in graph
	// 1. 2nd direction: q -> graph -> cluster -> t
	// matrix = max(matrix, min( w1, w2 = g(i,t))
	// also: use direct links via graph
	NeighbourList::const_iterator ei(neighbours[q].begin()), ei_end(neighbours[q].end());
      
	for(; ei != ei_end; ++ei)
	  {
	    Index i = edges[*ei].getNeighbour(q);
	    Weight w1 = edges[*ei].mWeight;
	    
	    if (param_loglevel >= 4) 
	      std::cout << "# graph: residue=" << map_vertex2token[i] << " weight=" << (int)w1 << std::endl;
	  
	    //----------------------------------------------------------	
	    // i is already sbjct sequence
	    // matrix = max(matrix, w1 = g(s,i))
	    if (map_vertex2prefix_index[i] == sbjct && weights[map_index2weights[i]] < w1)
	      weights[map_index2weights[i]] = w1;
	  
	    //----------------------------------------------------------
	    // iterate over all members t of cluster that i belongs to
	    // matrix = max(matrix, min( w1, w2 = h(i,t))
	    Index ic = map_vertex2cluster[i];
	    
	    Cluster::const_iterator t_it( clusters[ic].begin()), t_end( clusters[ic].end());
	    for (; t_it !=t_end; ++t_it)
	      {
		Index t = *t_it;
	      
		if (map_vertex2prefix_index[t] == sbjct)
		  {
		    Weight w2;
		    
		    if (i != t )
		      w2 = GetScoreCluster( tree, i, t);	
		    else
		      w2 = MAX_WEIGHT;

		    Weight w = std::min( w1, w2);
		    
		    if (param_loglevel >= 4) 
		      std::cout << "#   cluster: " << ic 
				<< " residue=" << map_vertex2token[t]
				<< " weight=" << (int)w2 			
				<< " sequence=" << map_prefix_index2prefix[map_vertex2prefix_index[t]]
				<< " total=" << (int)w
				<< std::endl;
		
		    if (weights[map_index2weights[t]] < w)
		      weights[map_index2weights[t]] = w;
		  }
	      }
	  }
      }
      
      //-----------------------------------------------------------------------
      // save dots
      { 
	Residue query_residue = GetResidue( map_vertex2token[q]);
	for (unsigned int i =0; i < weights.size(); ++i)
	  if (weights[i] >= param_min_dot_score)
	    {
	      Residue sbjct_residue = GetResidue(map_vertex2token[map_weights2vertex[i]]);
	      dots->addPair( query_residue, sbjct_residue, (CALCTYPE)weights[i]);
	    }
      }
      row++;

    } /* end of loop over residues in query */
  }
}

//------------------------------------------------------------
Alignandum * makeSequenceDummy( const ALIGNANDUM_POSITION length ) {

  char * buffer = new char[length + 1];
  for (int i = 0; i < length; ++i) buffer[i] = 'A';
  buffer[length] = '\0';
  Alignandum * result = makeSequence( buffer );
  delete [] buffer;

  return result;
}

//------------------------------------------------------------
void AlignDots( Alignata * dots, const char * prefix1 = NULL, const char * prefix2 = NULL)
{
  
  if (dots->getLength() == 0)
    return;

    
  Alignator * dottor = makeAlignatorDummy( dots );
  Alignator * alignator = makeAlignatorDotsSquared(param_gop, param_gep, dottor);
  
  Alignandum * seq1 = makeSequenceDummy( dots->getRowTo());
  Alignandum * seq2 = makeSequenceDummy( dots->getColTo());

  Alignata * result = makeAlignataVector();
  
  if (param_iterative_alignment > 0)
    {
      performIterativeAlignment( result, seq1, seq2, alignator, param_iterative_alignment );
    }
  else
    {
      alignator->Align( seq1, seq2, result);
    }

  AlignataIterator it = result->begin();
  AlignataIterator it_end = result->end();

  if (param_output_dots)
    {
      std::cout << "# score=" << result->getScore()
		<< " length=" << result->getLength()
		<< " gaps=" << result->getNumGaps()
		<< std::endl;
    
      if (prefix1 && prefix2)
	for (; it!=it_end; ++it)
	  std::cout << prefix1 << SEPARATOR_PREFIX << it->mRow << SEPARATOR
		    << prefix2 << SEPARATOR_PREFIX << it->mCol << SEPARATOR
		    << it->mScore << std::endl;
      else
	for (; it!=it_end; ++it) 
	  std::cout << it->mRow << SEPARATOR
		    << it->mCol << SEPARATOR
		    << it->mScore << std::endl;
    }

  if (param_output_pairs)
    {
      std::string row_ali, col_ali;
      
      writeAlignataCompressed( (const Alignata*)result, row_ali, col_ali );
      if (prefix1 && prefix2)
	std::cout << prefix1 << SEPARATOR 
		  << prefix2 << SEPARATOR;

      std::cout << result->getScore() << SEPARATOR
		<< result->getRowFrom() << SEPARATOR
		<< result->getRowTo() << SEPARATOR
		<< row_ali << SEPARATOR
		<< result->getColFrom() << SEPARATOR
		<< result->getColTo() << SEPARATOR
		<< col_ali << SEPARATOR
		<< result->getLength() << SEPARATOR
		<< result->getNumGaps() << SEPARATOR
		<< std::endl;
    }

  delete seq1;
  delete seq2;
  delete dottor;
  delete alignator;
  delete result;
}

//----------------------------------------------------------------------------------
void DoAlignment( const char * prefix_query,
		  const char * prefix_sbjct,
		  const Edges & edges,
		  const Neighbours & neighbours,
		  const Residues & residues,
		  const ClusterTree & tree,
		  const Clusters & clusters,
		  const MapVertex2Index & map_vertex2cluster,
		  const MapVertex2Token & map_vertex2token,
		  const MapPrefix2PrefixIndex & map_prefix2prefix_index,
		  const MapVertex2Index & map_vertex2prefix_index,
		  const MapPrefixIndex2Prefix & map_prefix_index2prefix) 
{
  
  Alignata * dots = makeAlignataMatrixRow();
  
  {
    time_t t1 = time(NULL);
    CalculateDots( dots,
		   edges, neighbours,
		   prefix_query, prefix_sbjct,
		   residues, tree, clusters,
		   map_vertex2cluster, map_vertex2token, map_prefix2prefix_index,
		   map_vertex2prefix_index, map_prefix_index2prefix);
    
    time_t t2 = time(NULL);
    
    if (param_loglevel >= 1) 
      std::cout << "# calculating dots: " << t2 - t1 << " seconds." << std::endl;
  }
  
  {
    time_t t1 = time(NULL);
    AlignDots( dots, prefix_query, prefix_sbjct );
    time_t t2 = time(NULL);
    
    if (param_loglevel >= 1) 
      std::cout << "# aligning dots: " << t2 - t1 << " seconds." << std::endl;
  }

  delete dots;
}

//------------------------------------------------------------
int main (int argc, char *argv[]) {

  ParseArguments( argc, argv );

  MapToken2Vertex map_token2vertex;
  MapVertex2Token map_vertex2token;

  //-----------------------------------------------------------------
  // read neighbours
  Neighbours neighbours;
  Edges edges;
    
  {
    time_t t1 = time(NULL);
    
    FILE * file;

    if (param_filename_links.length() > 0)
      file = fopen(param_filename_links.c_str(),"r");
    else 
      file = stdin;
  
    if (file == NULL) {
      std::cout << "# could not open input file " << param_filename_links << std::endl;
      exit(EXIT_FAILURE);
    }

    if (param_pairwise) 
      BuildNeighbourListsFromAlignments( edges, neighbours, map_token2vertex, map_vertex2token, file );
    else 
      BuildNeighbourListsFromDots( edges, neighbours, map_token2vertex, map_vertex2token, file );

    if (param_filename_links.length() > 0)
      fclose(file);

    time_t t2 = time(NULL);

    if (param_loglevel >= 1) 
      std::cout << "# reading graph: " << t2 - t1 << " s" << std::endl;
    
  }

  //-----------------------------------------------------------------
  // build prefix maps
  MapPrefix2PrefixIndex map_prefix2prefix_index;
  MapPrefixIndex2Prefix map_prefix_index2prefix;
  MapVertex2Index map_vertex2prefix_index;
  Residues residues;
  
  {
    BuildPrefixMaps( residues,
		     map_prefix2prefix_index,
		     map_prefix_index2prefix,
		     map_vertex2prefix_index,
		     map_vertex2token);
    
  }
  
  {
    if (param_loglevel >= 1)
      {
	// iterate over all edges
	Neighbours::iterator it(neighbours.begin()), end(neighbours.end());
	long nedges = 0;
	for (;it != end; ++it) nedges += it->size();
	
	std::cout << "# graph has " 
		  << map_vertex2token.size() << " vertices and " 
		  << nedges << " edges."
		  << std::endl;
      }

    
    PrintNeighbours( edges, neighbours, map_vertex2token );
  }

  //-----------------------------------------------------------------  
  // calculate edge weights
  {

    time_t t1 = time(NULL);
    CalculateEdgeWeights( edges, neighbours, map_vertex2token );    
    time_t t2 = time(NULL);
    
    if (param_loglevel >= 1) 
      std::cout << "# weighting edges: " << t2 - t1 << " seconds." << std::endl;

    // print weighted edges as a dot-list
    if (param_dump_data & 1)
      {
	Neighbours::const_iterator it(neighbours.begin()), end(neighbours.end());
	std::cout << "# weighted edges: start" << std::endl;
	for (Index i = 0; it != end; ++it, ++i)
	  {
	    if ((*it).size() == 1) continue;
	    NeighbourList::const_iterator iit( it->begin()), iend( it->end());
	    for (;iit!=iend;++iit)
	      {
		Vertex n = edges[*iit].getNeighbour(i);
		if (n < i) continue;
		std::cout << "#\t"
			  << map_vertex2token[i] << "\t"
			  << map_vertex2token[n] << "\t"
			  << (int)edges[*iit].mWeight
			  << std::endl;
	      }
	  }
	std::cout << "# weighted edges: end" << std::endl;	
      }
    
    PrintNeighbours( edges, neighbours, map_vertex2token );    
  }

  //-----------------------------------------------------------------
  // calculate clusters together with minimum spanning trees
  MapVertex2Index map_vertex2cluster;  
  Clusters clusters;
  ClusterTree tree( map_token2vertex.size());
  
  {

    time_t t1 = time(NULL);
    CalculateClusters( edges,
		       neighbours,
		       map_vertex2prefix_index,
		       map_vertex2cluster,
		       clusters,
		       tree);
    time_t t2 = time(NULL);
    
    if (param_loglevel >= 1) 
      std::cout << "# clustering: " << t2 - t1 << " seconds." << std::endl;
    

    if (param_loglevel >= 4)
      {
	std::cout << "# //" << std::endl;
	std::cout << "# assignments of residues to clusters" << std::endl;
	for (Vertex v = 0; v < (Vertex)map_vertex2token.size(); ++v) 
	  std::cout << "# " << v << "\t" << map_vertex2token[v] << "\t" << map_vertex2cluster[v] << std::endl;
      }

    if (param_loglevel >= 4)
      {
	std::cout << "# //" << std::endl;
	std::cout << "# cluster memberships" << std::endl;
	for (unsigned int i = 0; i < clusters.size(); ++i) 
	  std::cout << "# " << i << "\t" << clusters[i] << std::endl;
      }

#ifdef DEBUG
    {
      std::cout << "# sanity check: cluster sizes...";
      bool ok = true;
      for (unsigned int i = 0; i < clusters.size(); ++i)
	if (clusters[i].size() > map_prefix_index2prefix.size())
	  ok = false;
      if (ok) std::cout << "ok" << std::endl; else std::cout << "failed" << std::endl;
    }
#endif    

#ifdef DEBUG
    {
      std::cout << "# sanity check: cluster prefixes...";
      bool ok = true;
      for (unsigned int i = 0; i < clusters.size(); ++i)
	{
	  std::vector<Index> prefixes;

	  Cluster::const_iterator it(clusters[i].begin()), end(clusters[i].end());

	  for( ;it!=end;++it )
	    prefixes.push_back( map_vertex2prefix_index[*it]);
	  
	  std::sort(prefixes.begin(), prefixes.end());
	  
	  for (unsigned int x = 1; x < prefixes.size(); ++x)
	    if (prefixes[x-1] == prefixes[x])
	      ok = false;
	}
      
      if (ok) std::cout << "ok" << std::endl; else std::cout << "failed" << std::endl;
    }
#endif    
    
  }

  //-----------------------------------------------------------------
  // do the alignment

  //----------------------------------------------------------------------------
  // main loop over things to be aligned
  
  char * prefix_query = NULL;
  char * prefix_sbjct = NULL;
  unsigned int current_index = 0;        
  switch (param_alignment_mode)
    {
      // align in pairs
    case 0:
      while (current_index < param_prefixes.size())
	{
	  prefix_query = param_prefixes[current_index++];
	  prefix_sbjct = param_prefixes[current_index++];
	  DoAlignment( prefix_query, prefix_sbjct,
		       edges, neighbours, residues, tree, clusters,
		       map_vertex2cluster, map_vertex2token, map_prefix2prefix_index,
		       map_vertex2prefix_index, map_prefix_index2prefix);
	}
      break;
      // align first against all others
    case 1: 
      prefix_query = param_prefixes[current_index++];
      
      while (current_index < param_prefixes.size())
	{
	  prefix_sbjct = param_prefixes[current_index++];
	  DoAlignment( prefix_query, prefix_sbjct,
		       edges, neighbours, residues, tree, clusters,
		       map_vertex2cluster, map_vertex2token, map_prefix2prefix_index,
		       map_vertex2prefix_index, map_prefix_index2prefix);
	}
      break;
      // align all against all prefixes
    case 2:
      for (unsigned int i =0; i < param_prefixes.size(); ++i)
	{
	  prefix_query = param_prefixes[i];
	  for (unsigned int j = 0; j < param_prefixes.size() ; ++j)
	    {
	      if (j == i) continue;
	      prefix_sbjct = param_prefixes[j];
	      DoAlignment( prefix_query, prefix_sbjct,
			   edges, neighbours, residues, tree, clusters,
			   map_vertex2cluster, map_vertex2token, map_prefix2prefix_index,
			   map_vertex2prefix_index, map_prefix_index2prefix);
	    }
	}
      break;
      // align prefixes against all
    case 3:
      for (unsigned int i = 0; i < param_prefixes.size(); ++i)      
	{
	  prefix_query = param_prefixes[i];
	  for (unsigned int j = 0; j < map_prefix2prefix_index.size(); ++j)
	    {
	      prefix_sbjct = map_prefix_index2prefix[j];
	      if (strcmp(prefix_query, prefix_sbjct) == 0) continue;
	      
	      DoAlignment( prefix_query, prefix_sbjct,
			   edges, neighbours, residues, tree, clusters,
			   map_vertex2cluster, map_vertex2token, map_prefix2prefix_index,
			   map_vertex2prefix_index, map_prefix_index2prefix);
	    }
	}
      break;
      // align all against all
    case 4: 
      for (unsigned int i = 0; i< map_prefix2prefix_index.size() -1; ++i)
	for (unsigned int j = i+1; j < map_prefix2prefix_index.size(); ++j)
	  {

	    char * prefix_query = map_prefix_index2prefix[i];
	    char * prefix_sbjct = map_prefix_index2prefix[j];
	    
	    DoAlignment( prefix_query, prefix_sbjct,
			 edges, neighbours, residues, tree, clusters,
			 map_vertex2cluster, map_vertex2token, map_prefix2prefix_index,
			 map_vertex2prefix_index, map_prefix_index2prefix);
			 
	  }
      break;
    }
  
  // delete prefixes
  for (unsigned int i = 0; i < param_prefixes.size(); ++i)
    delete [] param_prefixes[i];

  if (param_loglevel >= 1)
    std::cout << "# maxflow finished successfully" << std::endl;
}












  
