/*
  maxflow - a program for sensitive sequence alignment

  $Id: maxflow.h,v 1.3 2004/02/24 09:50:11 aheger Exp $

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#ifndef MAXFLOW_H
#define MAXFLOW_H

#include <vector>
#include <map>
#include <set>
#include <list>
#include <cstdlib>
#include <cstdio>

// #define DEBUG 1
#define MAX_LINE_LENGTH 10000
#define MAX_TOKEN_SIZE 30
#define SEPARATOR '\t'
#define MAX_WEIGHT 255

namespace maxflow {

  typedef int Residue;
  typedef int Index;
  typedef int Vertex;
  typedef char * Token;
  typedef unsigned char Weight;
  typedef long Nid;
  
  struct ltstr
  {
    bool operator()(const char* s1, const char* s2) const
    {
      return strcmp(s1, s2) < 0;
    }
  };

  typedef std::map< Token, Index, ltstr > MapToken2Vertex;
  typedef std::vector< Token > MapVertex2Token;
  typedef std::vector< Index > MapVertex2Index;

  // definition of edge-properties => graph is weighted
  // note: the source is added here as well, because it is
  // needed for calculating the tree.
  struct Edge {
    Edge( const Index & s, const Index & t, const Weight & p = 1, const Weight & w = 0) :
      mSource(s), mTarget(t), mPriority(p), mWeight(w) {};
    Index mSource;
    Index mTarget;
    Weight mPriority;
    Weight mWeight;

    inline Index getNeighbour( Index & s) const { if (s == mSource) return mTarget; else return mSource; }
  };

  // as one of the ends is the same, compare the sum
  bool operator<=(const Edge & a, const Edge & b) { return a.mSource + a.mTarget <= b.mSource + b.mTarget; }  
  bool operator< (const Edge & a, const Edge & b) { return a.mSource + a.mTarget <  b.mSource + b.mTarget; }
  bool operator>=(const Edge & a, const Edge & b) { return a.mSource + a.mTarget >= b.mSource + b.mTarget; }
  bool operator> (const Edge & a, const Edge & b) { return a.mSource + a.mTarget >  b.mSource + b.mTarget; }
  bool operator==(const Edge & a, const Edge & b) { return a.mSource + a.mTarget == b.mSource + b.mTarget; }
  bool operator!=(const Edge & a, const Edge & b) { return a.mSource + a.mTarget != b.mSource + b.mTarget; }    
  
  std::ostream & operator<<( std::ostream & output, const Edge & src) {
    output << src.mSource << "-" << src.mTarget << "(" << (int)src.mWeight << ")";    
    return output;
  }  

  typedef long EdgeIndex;
  typedef std::vector< Edge > Edges;
  typedef std::vector< EdgeIndex > EdgeVector;
  typedef std::list< EdgeIndex > NeighbourList;
  typedef std::vector< NeighbourList > Neighbours;

  std::ostream & operator<<( std::ostream & output, const NeighbourList & src) {
    std::copy( src.begin(), src.end(), std::ostream_iterator< EdgeIndex>( output, "-"));
    return output;
  }  

  
  typedef std::map< Token, Index, ltstr > MapPrefix2PrefixIndex;
  typedef std::vector< Token > MapPrefixIndex2Prefix;
  typedef std::set<Index> PrefixSet;

  // clusters
  typedef std::list<Index> Cluster;
  typedef std::vector<Cluster> Clusters;

  std::ostream & operator<<( std::ostream & output, const Cluster & src) {
    std::copy( src.begin(), src.end(), std::ostream_iterator< Vertex >( output, " "));
    return output;
  }  
  
  // residue lists
  typedef std::list<Index> ResidueList;
  typedef std::vector<ResidueList> Residues;

//----------------------------------------------------------------------------------
// template functions

/*
 */
template< class T>
inline void PrintIndex( T & index, const char * title = "" ) {
  typedef typename T::iterator iterator;
  std::cout << "# ";
  unsigned int i = 0;
  iterator it(index.begin()), end(index.end());

  std::cout << title;

  for (; it != end; ++ it, ++i) 
    std::cout << i << "->" << *it << ":";
  std::cout << std::endl;
}



template<class A2I, class I2A>
typename A2I::data_type UpdateMaps( A2I & map_a2i, 
				    I2A & map_i2a, 
				    const typename A2I::key_type & v) {

  typedef typename A2I::iterator iterator;
  typedef typename I2A::size_type size_type;
  
  iterator it;
  size_type i;

  if ( (it = map_a2i.find(v)) == map_a2i.end()) {
    i = map_i2a.size();
    map_a2i[v] = i;
    map_i2a.push_back(v);
  } else {
    i = (*it).second;
  }

  return i;
}  

template<class A2I, class I2A>
typename A2I::data_type UpdateMapsDelete( A2I & map_a2i, 
					  I2A & map_i2a, 
					  const typename A2I::key_type & v) {

  typedef typename A2I::iterator iterator;
  typedef typename I2A::size_type size_type;
  
  iterator it;
  size_type i;

  if ( (it = map_a2i.find(v)) == map_a2i.end()) {
    i = map_i2a.size();
    map_a2i[v] = i;
    map_i2a.push_back(v);
  } else {
    i = (*it).second;
    delete [] v;
  }

  return i;
}  

}





#endif
