/*
  maxflow - a program for sensitive sequence alignment

  $Id: prefix.h,v 1.2 2004/02/10 09:34:44 aheger Exp $

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#ifndef __PREFIX_H
#define PREFIX_H

namespace MaxFlow {

#include <iosfwd>

template<class T, class W>
struct Node {
  Node() : mParent(0), mLeft(0), mRight(0), mWeight(0), mLevel(0) {};
  Node(const T & p, const T & l, const T & r, const W & w = 0, const unsigned int & e = 0) : 
    mParent(p), mLeft(l), mRight(r), mWeight(w), mLevel(e) {};
  T mParent, mLeft, mRight;
  W mWeight;
  unsigned int mLevel;
};

template<class T, class W>
std::ostream & operator<<( std::ostream & output, const Node<T, W> & src) {
  output << src.mParent << "\t" << src.mLeft << "\t" << src.mRight << "\t" << src.mWeight << "\t" << src.mLevel;
  return output;
}

template<class T, class W>
class Tree {
  friend std::ostream & operator<< <T,W>( std::ostream & output, const Tree<T, W> & src);

  typedef std::vector< Node<T,W> > NodeVector;
 public:
  Tree( T size) : mLastNode(size) { mNodes.resize(2 * size + 1); }

  /** join two nodes in the tree */
  T joinNodes( const T & a, const T & b, const W & w) {
    
    mNodes[a].mParent = mLastNode;
    mNodes[b].mParent = mLastNode;

    mNodes[mLastNode].mLevel = std::max(mNodes[a].mLevel, mNodes[b].mLevel) + 1;

    mNodes[mLastNode].mLeft = a;
    mNodes[mLastNode].mRight = b;
    mNodes[mLastNode].mWeight = w;
    
    return mLastNode++;
  }
  
  T getParent( const T & a) const { return mNodes[a].mParent; };
  T getLeft( const T & a) const { return mNodes[a].mLeft; };
  T getRight( const T & a) const { return mNodes[a].mRight; };
  W getWeight( const T & a) const { return mNodes[a].mWeight; };
  T isRoot( const T & a) const { return mNodes[a].mParent == 0; }; // parents have index > mNumLeaves
  T isLeaf( const T & a) const { return mNodes[a].mLeft == mNodes[a].mRight; };
  T getLastParent( const T & a) const { T x = a; while (!isRoot(x)) x = mNodes[x].mParent; return x;};
  unsigned int getLevel( const T & a) const { return mNodes[a].mLevel; }

  typename NodeVector::size_type size() const { return mNodes.size(); }

 private:
  NodeVector mNodes;
  T mLastNode;
};


template< class T, class W>
  std::ostream & operator<<( std::ostream & output, const Tree<T,W> & src) {
  for (unsigned int x = 0; x < src.mNodes.size(); ++x)
    output << x << "\t" << src.mNodes[x] << std::endl;
  return output;
}



} // namespace
#endif


